from typing import Union

import numpy as np

from jesse.helpers import get_candle_source, slice_candles


def linearreg_slope(candles: np.ndarray, period: int = 14, source_type: str = "close", sequential: bool = False) -> \
        Union[float, np.ndarray]:
    """
    LINEARREG_SLOPE - Linear Regression Slope

    :param candles: np.ndarray
    :param period: int - default: 14
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    if len(candles.shape) == 1:
        source = candles
    else:
        candles = slice_candles(candles, sequential)
        source = get_candle_source(candles, source_type=source_type)

    n = len(source)
    result = np.full(n, np.nan, dtype=float)
    if n < period:
        return result if sequential else result[-1]

    # Create a constant x-axis for the regression
    X = np.arange(period, dtype=float)
    sumX = X.sum()
    sumX2 = (X**2).sum()
    denom = period * sumX2 - sumX**2

    # Compute the rolling window slopes using vectorized operations
    windows = np.lib.stride_tricks.sliding_window_view(source, window_shape=period)
    sum_y = windows.sum(axis=1)
    sum_xy = (windows * X).sum(axis=1)
    slopes = (period * sum_xy - sumX * sum_y) / denom

    result[period-1:] = slopes
    
    return result if sequential else result[-1]



from setuptools import setup, find_packages

# also change in version.py
VERSION = "1.10.7"
DESCRIPTION = "A trading framework for cryptocurrencies"
with open("requirements.txt", "r", encoding="utf-8") as f:
    REQUIRED_PACKAGES = f.read().splitlines()

with open("README.md", "r", encoding="utf-8") as f:
    LONG_DESCRIPTION = f.read()

setup(
    name='jesse',
    version=VERSION,
    author="Saleh Mir",
    author_email="saleh@jesse.trade",
    packages=find_packages(),
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://jesse.trade",
    project_urls={
        'Documentation': 'https://docs.jesse.trade',
        'Say Thanks!': 'https://jesse.trade/discord',
        'Source': 'https://github.com/jesse-ai/jesse',
        'Tracker': 'https://github.com/jesse-ai/jesse/issues',
    },
    install_requires=REQUIRED_PACKAGES,
    entry_points='''
        [console_scripts]
        jesse=jesse.__init__:cli
    ''',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.10',
    include_package_data=True,
    package_data={
        '': ['*.dll', '*.dylib', '*.so', '*.json'],
    },
)



import json
from starlette.responses import JSONResponse
import jesse.helpers as jh
from jesse.services import transformers


def get_notification_api_keys() -> JSONResponse:
    from jesse.services.db import database
    database.open_connection()

    from jesse.models.NotificationApiKeys import NotificationApiKeys

    try:
        # fetch all the notification api keys
        api_keys = NotificationApiKeys.select()
    except Exception as e:
        database.close_connection()
        return JSONResponse({
            'status': 'error',
            'message': str(e)
        }, status_code=500)

    # transform each api_key using transformers.get_notification_api_key()
    api_keys = [transformers.get_notification_api_key(api_key) for api_key in api_keys]

    database.close_connection()

    return JSONResponse({
        'data': api_keys
    }, status_code=200)


def store_notification_api_keys(
        name: str,
        driver: str,
        fields: dict
) -> JSONResponse:
    from jesse.services.db import database
    database.open_connection()

    from jesse.models.NotificationApiKeys import NotificationApiKeys

    # check if the api key already exists
    if NotificationApiKeys.select().where(NotificationApiKeys.name == name).exists():
        database.close_connection()
        return JSONResponse({
            'status': 'error',
            'message': f'API key for the name "{name}" already exists. Please choose another driver.'
        }, status_code=400)

    try:
        # create the record
        notification_api_key: NotificationApiKeys = NotificationApiKeys.create(
            id=jh.generate_unique_id(),
            name=name,
            driver=driver,
            fields=json.dumps(fields),
            created_at=jh.now_to_datetime()
        )
    except ValueError as e:
        database.close_connection()
        return JSONResponse({
            'status': 'error',
            'message': str(e)
        }, status_code=400)
    except Exception as e:
        database.close_connection()
        return JSONResponse({
            'status': 'error',
            'message': str(e)
        }, status_code=500)

    database.close_connection()

    return JSONResponse({
        'status': 'success',
        'message': 'Notification API key has been stored successfully.',
        'data': transformers.get_notification_api_key(notification_api_key)
    }, status_code=200)


def delete_notification_api_keys(notification_api_key_id: str) -> JSONResponse:
    from jesse.services.db import database
    database.open_connection()

    from jesse.models.NotificationApiKeys import NotificationApiKeys

    try:
        # delete the record
        NotificationApiKeys.delete().where(NotificationApiKeys.id == notification_api_key_id).execute()
    except Exception as e:
        database.close_connection()
        return JSONResponse({
            'status': 'error',
            'message': str(e)
        }, status_code=500)

    database.close_connection()

    return JSONResponse({
        'status': 'success',
        'message': 'Notification API key has been deleted successfully.'
    }, status_code=200)



from typing import Union

import numpy as np
from numba import njit

from jesse.helpers import get_candle_source, same_length, slice_candles


def cg(candles: np.ndarray, period: int = 10, source_type: str = "close", sequential: bool = False) -> Union[
    float, np.ndarray]:
    """
    Center of Gravity (CG)

    :param candles: np.ndarray
    :param period: int - default: 10
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)

    source = get_candle_source(candles, source_type=source_type)
    res = go_fast(source, period)

    return same_length(candles, res) if sequential else res[-1]


@njit(cache=True)
def go_fast(source, period):  # Function is compiled to machine code when called the first time
    res = np.full_like(source, fill_value=np.nan)
    for i in range(source.size):
        if i > period:
            num = 0
            denom = 0
            for count in range(period - 1):
                close = source[i - count]
                if not np.isnan(close):
                    num += (1 + count) * close
                    denom += close
            result = -num / denom if denom != 0 else 0
            res[i] = result
    return res



from typing import Union
import numpy as np
from jesse.helpers import get_candle_source, same_length, slice_candles
from numba import njit


@njit
def linear_regression_line(x, y):
    n = len(x)
    sum_x = np.sum(x)
    sum_y = np.sum(y)
    sum_xy = np.sum(x * y)
    sum_xx = np.sum(x * x)

    slope = (n * sum_xy - sum_x * sum_y) / (n * sum_xx - sum_x * sum_x)
    intercept = (sum_y - slope * sum_x) / n

    return slope * x + intercept


def fosc(candles: np.ndarray, period: int = 5, source_type: str = "close", sequential: bool = False) -> Union[float, np.ndarray]:
    """
    FOSC - Forecast Oscillator

    :param candles: np.ndarray
    :param period: int - default: 5
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)
    source = get_candle_source(candles, source_type=source_type)

    # Prepare the result array
    res = np.zeros_like(source)

    # Calculate FOSC for each window
    for i in range(period - 1, len(source)):
        window = source[i - period + 1:i + 1]
        x = np.arange(period)
        predicted = linear_regression_line(x, window)
        res[i] = 100 * (window[-1] - predicted[-1]) / window[-1]

    # Replace initial NaN values with 0
    res[:period - 1] = 0

    return same_length(candles, res) if sequential else res[-1]



from jesse.strategies import Strategy
import jesse.helpers as jh


class TestEntryOrdersAndExitOrdersProperties(Strategy):
    def should_long(self):
        return self.price in [10, 25, 30]

    def go_long(self):
        if self.price == 10:
            self.buy = [
                (1, 10),
                (2, 9),
                (3, 8),
            ]
            self.stop_loss = 1, 5
        elif self.price == 25:
            self.buy = 1, 28
        elif self.price == 30:
            self.buy = [
                # stop
                (1, 31),
                # market
                (1, 30),
                # limit
                (1, 29),
            ]
            self.take_profit = [
                (1, 35),
                (1, 36),
                (1, 37),
            ]
            self.stop_loss = 3, 25

    def should_cancel_entry(self):
        # cancel second position entry attempt
        if self.price == 27:
            return True

        return False

    def update_position(self) -> None:
        if self.price == 20:
            self.liquidate()

    def before(self) -> None:
        if self.price == 11:
            # entry orders
            assert len(self.entry_orders) == 3
            assert self.entry_orders[0].price == 10
            assert self.entry_orders[0].qty == 1
            assert self.entry_orders[0].is_executed
            assert self.entry_orders[1].price == 9
            assert self.entry_orders[1].qty == 2
            assert self.entry_orders[1].is_active
            assert self.entry_orders[2].price == 8
            assert self.entry_orders[2].qty == 3
            assert self.entry_orders[2].is_active
            # exit orders
            assert len(self.exit_orders) == 1
            assert self.exit_orders[0].price == 5
            assert self.exit_orders[0].qty == -1
            assert self.exit_orders[0].is_active
            assert self.exit_orders[0].type == 'STOP'

        if self.price == 21:
            # entry orders
            assert len(self.entry_orders) == 0
            # exit orders
            assert len(self.exit_orders) == 0

        # after entry orders are submitted for the SECOND time
        if self.price == 26:
            # entry orders
            assert len(self.entry_orders) == 1
            assert self.entry_orders[0].price == 28
            assert self.entry_orders[0].qty == 1
            assert self.entry_orders[0].is_active
            assert self.entry_orders[0].type == 'STOP'
            # exit orders
            assert len(self.exit_orders) == 0

        if self.price == 28:
            assert len(self.entry_orders) == 0
            assert len(self.exit_orders) == 0

        # TODO: test for a third time while exit orders are also submitted
        if self.price == 31:
            # entry orders
            assert len(self.entry_orders) == 3
            assert self.entry_orders[0].price == 31
            assert self.entry_orders[0].qty == 1
            assert self.entry_orders[0].is_executed
            assert self.entry_orders[1].price == 30
            assert self.entry_orders[1].qty == 1
            assert self.entry_orders[1].is_executed
            assert self.entry_orders[2].price == 29
            assert self.entry_orders[2].qty == 1
            assert self.entry_orders[2].is_active

            # exit orders
            assert len(self.exit_orders) == 4

            assert self.exit_orders[0].price == 25
            assert self.exit_orders[0].qty == -3
            assert self.exit_orders[0].is_active
            assert self.exit_orders[0].type == 'STOP'

            assert self.exit_orders[1].price == 35
            assert self.exit_orders[1].qty == -1
            assert self.exit_orders[1].is_active
            assert self.exit_orders[1].type == 'LIMIT'

            assert self.exit_orders[2].price == 36
            assert self.exit_orders[2].qty == -1
            assert self.exit_orders[2].is_active
            assert self.exit_orders[2].type == 'LIMIT'

            assert self.exit_orders[3].price == 37
            assert self.exit_orders[3].qty == -1
            assert self.exit_orders[3].is_active
            assert self.exit_orders[3].type == 'LIMIT'
            



from datetime import datetime
from typing import List

import numpy as np
import pandas as pd

import jesse.helpers as jh
from jesse.models import ClosedTrade
from jesse.services import selectors
from jesse.store import store


def candles_info(candles_array: np.ndarray) -> dict:
    period = jh.date_diff_in_days(
        jh.timestamp_to_arrow(candles_array[0][0]),
        jh.timestamp_to_arrow(candles_array[-1][0])) + 1

    if period > 365:
        duration = f'{period} days ({round(period / 365, 2)} years)'
    elif period > 30:
        duration = f'{period} days ({round(period / 30, 2)} months)'
    else:
        duration = f'{period} days'

    # type of the exchange
    trading_exchange = selectors.get_trading_exchange()

    info = {
        'duration': duration,
        'starting_time': candles_array[0][0],
        'finishing_time': (candles_array[-1][0] + 60_000),
        'exchange_type': trading_exchange.type,
        'exchange': trading_exchange.name,
    }

    # if the exchange type is futures, also display leverage
    if trading_exchange.type == 'futures':
        info['leverage'] = trading_exchange.futures_leverage
        info['leverage_mode'] = trading_exchange.futures_leverage_mode

    return info


def routes(routes_arr: list) -> list:
    return [{
            'exchange': r.exchange,
            'symbol': r.symbol,
            'timeframe': r.timeframe,
            'strategy_name': r.strategy_name,
        } for r in routes_arr]


def _prepare_returns(returns, rf=0.0, periods=252):
    """
    Helper function to prepare returns data by converting to pandas Series and 
    adjusting for risk-free rate if provided
    """
    if rf != 0:
        returns = returns - (rf / periods)
    
    if isinstance(returns, pd.DataFrame):
        returns = returns[returns.columns[0]]
        
    return returns

def sharpe_ratio(returns, rf=0.0, periods=365, annualize=True, smart=False):
    """
    Calculates the sharpe ratio of access returns
    """
    returns = _prepare_returns(returns, rf, periods)
    divisor = returns.std(ddof=1)
    
    if smart:
        divisor = divisor * autocorr_penalty(returns)
        
    res = returns.mean() / divisor
    
    if annualize:
        res = res * np.sqrt(1 if periods is None else periods)
    
    # Always convert to pandas Series    
    return pd.Series([res])


def sortino_ratio(returns, rf=0, periods=365, annualize=True, smart=False):
    """
    Calculates the sortino ratio of access returns
    """
    returns = _prepare_returns(returns, rf, periods)
    
    downside = np.sqrt((returns[returns < 0] ** 2).sum() / len(returns))
    
    # Handle division by zero
    if downside == 0:
        res = np.inf if returns.mean() > 0 else -np.inf
    else:
        if smart:
            downside = downside * autocorr_penalty(returns)
            
        res = returns.mean() / downside
        
        if annualize:
            res = res * np.sqrt(1 if periods is None else periods)
    
    # Always convert to pandas Series
    return pd.Series([res])


def autocorr_penalty(returns):
    """
    Calculates the autocorrelation penalty for returns
    """
    num = len(returns)
    coef = np.abs(np.corrcoef(returns[:-1], returns[1:])[0, 1])
    corr = [((num - x) / num) * coef**x for x in range(1, num)]
    return np.sqrt(1 + 2 * np.sum(corr))


def calmar_ratio(returns):
    """
    Calculates the calmar ratio (CAGR% / MaxDD%)
    """
    # Get daily returns
    returns = _prepare_returns(returns)
    
    # Calculate CAGR exactly as in cagr() function
    first_value = 1
    last_value = (1 + returns).prod()
    days = (returns.index[-1] - returns.index[0]).days
    years = float(days) / 365
    
    if years == 0:
        return pd.Series([0.0])
        
    cagr_ratio = (last_value / first_value) ** (1 / years) - 1
    
    # Calculate Max Drawdown using cumulative returns
    cum_returns = (1 + returns).cumprod()
    rolling_max = cum_returns.expanding(min_periods=1).max()
    drawdown = cum_returns / rolling_max - 1
    max_dd = abs(drawdown.min())
    
    # Calculate Calmar
    result = cagr_ratio / max_dd if max_dd != 0 else 0
    
    # Always convert to pandas Series
    return pd.Series([result])


def max_drawdown(returns):
    """
    Calculates the maximum drawdown
    """
    prices = (returns + 1).cumprod()
    result = (prices / prices.expanding(min_periods=0).max()).min() - 1
    
    # Always convert to pandas Series
    return pd.Series([result])


def cagr(returns, rf=0.0, compounded=True, periods=365):
    """
    Calculates the communicative annualized growth return (CAGR%)
    """
    returns = _prepare_returns(returns, rf)
    
    # Get first and last values of cumulative returns
    first_value = 1
    last_value = (1 + returns).prod()
    
    # Calculate years exactly as quantstats does
    days = (returns.index[-1] - returns.index[0]).days
    years = float(days) / 365
    
    # Handle edge case
    if years == 0:
        return pd.Series([0.0])
        
    # Calculate CAGR using quantstats formula
    result = (last_value / first_value) ** (1 / years) - 1
    
    return pd.Series([result])


def omega_ratio(returns, rf=0.0, required_return=0.0, periods=365):
    """
    Determines the Omega ratio of a strategy
    """
    returns = _prepare_returns(returns, rf, periods)
    
    if periods == 1:
        return_threshold = required_return
    else:
        return_threshold = (1 + required_return) ** (1.0 / periods) - 1
        
    returns_less_thresh = returns - return_threshold
    numer = returns_less_thresh[returns_less_thresh > 0.0].sum()
    denom = -1.0 * returns_less_thresh[returns_less_thresh < 0.0].sum()
    
    result = numer / denom if denom > 0.0 else np.nan
    
    # Always convert to pandas Series
    return pd.Series([result])


def serenity_index(returns, rf=0):
    """
    Calculates the serenity index score
    """
    dd = to_drawdown_series(returns)
    pitfall = -conditional_value_at_risk(dd) / returns.std()
    result = (returns.sum() - rf) / (ulcer_index(returns) * pitfall)
    
    # Always convert to pandas Series
    return pd.Series([result])


def ulcer_index(returns):
    """
    Calculates the ulcer index score (downside risk measurement)
    """
    dd = to_drawdown_series(returns)
    return np.sqrt(np.divide((dd**2).sum(), returns.shape[0] - 1))


def to_drawdown_series(returns):
    """
    Convert returns series to drawdown series
    """
    prices = (1 + returns).cumprod()
    dd = prices / np.maximum.accumulate(prices) - 1.0
    return dd.replace([np.inf, -np.inf, -0], 0)


def conditional_value_at_risk(returns, sigma=1, confidence=0.95):
    """
    Calculates the conditional daily value-at-risk (aka expected shortfall)
    """
    if len(returns) < 2:
        return 0
        
    returns = _prepare_returns(returns)
    # Sort returns from worst to best
    sorted_returns = np.sort(returns)
    # Find the index based on confidence level
    index = int((1 - confidence) * len(sorted_returns))
    
    # Handle empty slice warning
    if index == 0:
        return sorted_returns[0] if len(sorted_returns) > 0 else 0
        
    # Calculate CVaR as the mean of worst losses
    c_var = sorted_returns[:index].mean()
    return c_var if ~np.isnan(c_var) else 0


def trades(trades_list: List[ClosedTrade], daily_balance: list, final: bool = True) -> dict:
    starting_balance = 0
    current_balance = 0

    for e in store.exchanges.storage:
        starting_balance += store.exchanges.storage[e].starting_assets[jh.app_currency()]
        current_balance += store.exchanges.storage[e].assets[jh.app_currency()]

    if not trades_list:
        return {'total': 0, 'win_rate': 0, 'net_profit_percentage': 0}

    df = pd.DataFrame.from_records([t.to_dict for t in trades_list])

    total_completed = len(df)
    winning_trades = df.loc[df['PNL'] > 0]
    total_winning_trades = len(winning_trades)
    losing_trades = df.loc[df['PNL'] < 0]
    total_losing_trades = len(losing_trades)

    arr = df['PNL'].to_numpy()
    pos = np.clip(arr, 0, 1).astype(bool).cumsum()
    neg = np.clip(arr, -1, 0).astype(bool).cumsum()
    current_streak = np.where(arr >= 0, pos - np.maximum.accumulate(np.where(arr <= 0, pos, 0)),
                              -neg + np.maximum.accumulate(np.where(arr >= 0, neg, 0)))

    s_min = current_streak.min()
    losing_streak = 0 if s_min > 0 else abs(s_min)

    s_max = current_streak.max()
    winning_streak = max(s_max, 0)

    largest_losing_trade = 0 if total_losing_trades == 0 else losing_trades['PNL'].min()
    largest_winning_trade = 0 if total_winning_trades == 0 else winning_trades['PNL'].max()
    if len(winning_trades) == 0:
        win_rate = 0
    else:
        win_rate = len(winning_trades) / (len(losing_trades) + len(winning_trades))
    longs_count = len(df.loc[df['type'] == 'long'])
    shorts_count = len(df.loc[df['type'] == 'short'])
    longs_percentage = longs_count / (longs_count + shorts_count) * 100
    shorts_percentage = 100 - longs_percentage
    fee = df['fee'].sum()
    net_profit = df['PNL'].sum()
    net_profit_percentage = (net_profit / starting_balance) * 100
    average_win = winning_trades['PNL'].mean()
    average_loss = abs(losing_trades['PNL'].mean())
    ratio_avg_win_loss = average_win / average_loss
    expectancy = (0 if np.isnan(average_win) else average_win) * win_rate - (
        0 if np.isnan(average_loss) else average_loss) * (1 - win_rate)
    expectancy = expectancy
    expectancy_percentage = (expectancy / starting_balance) * 100
    expected_net_profit_every_100_trades = expectancy_percentage * 100
    average_holding_period = df['holding_period'].mean()
    average_winning_holding_period = winning_trades['holding_period'].mean()
    average_losing_holding_period = losing_trades['holding_period'].mean()
    gross_profit = winning_trades['PNL'].sum()
    gross_loss = losing_trades['PNL'].sum()

    start_date = datetime.fromtimestamp(store.app.starting_time / 1000)
    date_index = pd.date_range(start=start_date, periods=len(daily_balance))

    daily_return = pd.DataFrame(daily_balance, index=date_index).pct_change(1)

    total_open_trades = store.app.total_open_trades
    open_pl = store.app.total_open_pl

    # Helper function to safely convert values
    def safe_convert(value, convert_type=float):
        try:
            if isinstance(value, pd.Series):
                value = value.iloc[0]
            if np.isnan(value):
                return np.nan
            return convert_type(value)
        except:
            return np.nan

    # Calculate metrics using 365 days for crypto markets
    max_dd = np.nan if len(daily_return) < 2 else max_drawdown(daily_return).iloc[0] * 100
    annual_return = np.nan if len(daily_return) < 2 else cagr(daily_return, periods=365).iloc[0] * 100
    sharpe = np.nan if len(daily_return) < 2 else sharpe_ratio(daily_return, periods=365).iloc[0]
    calmar = np.nan if len(daily_return) < 2 else calmar_ratio(daily_return).iloc[0]
    sortino = np.nan if len(daily_return) < 2 else sortino_ratio(daily_return, periods=365).iloc[0]
    omega = np.nan if len(daily_return) < 2 else omega_ratio(daily_return, periods=365).iloc[0]
    serenity = np.nan if len(daily_return) < 2 else serenity_index(daily_return).iloc[0]

    return {
        'total': safe_convert(total_completed, int),
        'total_winning_trades': safe_convert(total_winning_trades, int),
        'total_losing_trades': safe_convert(total_losing_trades, int),
        'starting_balance': safe_convert(starting_balance),
        'finishing_balance': safe_convert(current_balance),
        'win_rate': safe_convert(win_rate),
        'ratio_avg_win_loss': safe_convert(ratio_avg_win_loss),
        'longs_count': safe_convert(longs_count, int),
        'longs_percentage': safe_convert(longs_percentage),
        'shorts_percentage': safe_convert(shorts_percentage),
        'shorts_count': safe_convert(shorts_count, int),
        'fee': safe_convert(fee),
        'net_profit': safe_convert(net_profit),
        'net_profit_percentage': safe_convert(net_profit_percentage),
        'average_win': safe_convert(average_win),
        'average_loss': safe_convert(average_loss),
        'expectancy': safe_convert(expectancy),
        'expectancy_percentage': safe_convert(expectancy_percentage),
        'expected_net_profit_every_100_trades': safe_convert(expected_net_profit_every_100_trades),
        'average_holding_period': safe_convert(average_holding_period),
        'average_winning_holding_period': safe_convert(average_winning_holding_period),
        'average_losing_holding_period': safe_convert(average_losing_holding_period),
        'gross_profit': safe_convert(gross_profit),
        'gross_loss': safe_convert(gross_loss),
        'max_drawdown': safe_convert(max_dd),
        'annual_return': safe_convert(annual_return),
        'sharpe_ratio': safe_convert(sharpe),
        'calmar_ratio': safe_convert(calmar),
        'sortino_ratio': safe_convert(sortino),
        'omega_ratio': safe_convert(omega),
        'serenity_index': safe_convert(serenity),
        'total_open_trades': safe_convert(total_open_trades, int),
        'open_pl': safe_convert(open_pl),
        'winning_streak': safe_convert(winning_streak, int),
        'losing_streak': safe_convert(losing_streak, int),
        'largest_losing_trade': safe_convert(largest_losing_trade),
        'largest_winning_trade': safe_convert(largest_winning_trade),
        'current_streak': safe_convert(current_streak[-1], int),
    }


def hyperparameters(routes_arr: list) -> list:
    if routes_arr[0].strategy.hp is None:
        return []
    # only for the first route
    hp = []

    # add DNA
    dna_value = routes_arr[0].strategy.dna()
    if dna_value is not None and len(dna_value) > 16:
        formatted_dna = f"{dna_value[:5]}*****{dna_value[-5:]}"
        hp.append(['DNA', formatted_dna])
    else:
        hp.append(['DNA', dna_value])

    # add hyperparameters
    for key in routes_arr[0].strategy.hp:
        hp.append([
            key, routes_arr[0].strategy.hp[key]
        ])
    return hp



import threading
from typing import Union

import jesse.helpers as jh
from jesse.models import Order
from jesse.services import logger


class API:
    def __init__(self) -> None:
        self.drivers = {}

        if not jh.is_live():
            self.initiate_drivers()

    def initiate_drivers(self) -> None:
        considering_exchanges = jh.get_config('app.considering_exchanges')

        # A helpful assertion
        if not len(considering_exchanges):
            raise Exception('No exchange is available for initiating in the API class')

        for e in considering_exchanges:
            if jh.is_live():
                def initiate_ws(exchange_name: str) -> None:
                    exchange_class = jh.get_config(f'app.live_drivers.{exchange_name}')
                    self.drivers[exchange_name] = exchange_class()

                threading.Thread(target=initiate_ws, args=[e]).start()
            else:
                from jesse.exchanges import Sandbox
                self.drivers[e] = Sandbox(e)

    def market_order(
        self,
        exchange: str,
        symbol: str,
        qty: float,
        current_price: float,
        side: str,
        reduce_only: bool
    ) -> Union[Order, None]:
        if exchange not in self.drivers:
            logger.info(f'Exchange "{exchange}" driver not initiated yet. Trying again in the next candle')
            return None
        return self.drivers[exchange].market_order(symbol, qty, current_price, side, reduce_only)

    def limit_order(
        self,
        exchange: str,
        symbol: str,
        qty: float,
        price: float,
        side: str,
        reduce_only: bool
    ) -> Union[Order, None]:
        if exchange not in self.drivers:
            logger.info(f'Exchange "{exchange}" driver not initiated yet. Trying again in the next candle')
            return None
        return self.drivers[exchange].limit_order(symbol, qty, price, side, reduce_only)

    def stop_order(
        self, exchange: str,
        symbol: str,
        qty: float,
        price: float,
        side: str,
        reduce_only: bool
    ) -> Union[Order, None]:
        if exchange not in self.drivers:
            logger.info(f'Exchange "{exchange}" driver not initiated yet. Trying again in the next candle')
            return None
        return self.drivers[exchange].stop_order(symbol, qty, price, side, reduce_only)

    def cancel_all_orders(self, exchange: str, symbol: str) -> bool:
        if exchange not in self.drivers:
            logger.info(f'Exchange "{exchange}" driver not initiated yet. Trying again in the next candle')
            return False
        return self.drivers[exchange].cancel_all_orders(symbol)

    def cancel_order(self, exchange: str, symbol: str, order_id: str) -> bool:
        if exchange not in self.drivers:
            logger.info(f'Exchange "{exchange}" driver not initiated yet. Trying again in the next candle')
            return False
        return self.drivers[exchange].cancel_order(symbol, order_id)


api = API()



import requests
from fastapi.responses import JSONResponse
from jesse.services.auth import get_access_token
import jesse.helpers as jh
import json
from jesse.info import JESSE_API_URL


def feedback(description: str, email: str = None) -> JSONResponse:
    access_token = get_access_token()

    res = requests.post(
        JESSE_API_URL + '/feedback', {
            'description': description,
            'email': email
        },
        headers={'Authorization': f'Bearer {access_token}'}
    )

    success_message = 'Feedback submitted successfully'
    error_message = f"{res.status_code} error: {res.json()['message']}"

    return JSONResponse({
        'status': 'success' if res.status_code == 200 else 'error',
        'message': success_message if res.status_code == 200 else error_message
    }, status_code=200)


def report_exception(
        description: str, traceback: str, mode: str, attach_logs: bool, session_id: str, email: str = None, has_live: bool = False
) -> JSONResponse:
    access_token = get_access_token()

    if attach_logs and session_id:
        path_exchange_log = None
        if mode == 'backtest':
            path_log = f'storage/logs/backtest-mode/{session_id}.txt'
        elif mode == 'live':
            path_log = f'storage/logs/live-mode/{session_id}.txt'
            path_exchange_log = f'storage/logs/live-mode/{session_id}-raw-exchange-logs.txt'
        else:
            raise ValueError('Invalid mode')

        # attach exchange_log if there's any
        files = {'log_file': open(path_log, 'rb')}
        if path_exchange_log and jh.file_exists(path_exchange_log):
            files['exchange_log'] = open(path_exchange_log, 'rb')
    else:
        files = None

    from jesse.version import __version__ as jesse_version
    info = {
        'os': jh.get_os(),
        'python_version': '{}.{}'.format(*jh.python_version()),
        'is_docker': jh.is_docker(),
        'jesse_version': jesse_version
    }
    if has_live:
        from jesse_live.version import __version__ as live_plugin_version
        info['live_plugin_version'] = live_plugin_version

    params = {
        'description': description,
        'traceback': traceback,
        'email': email,
        'info': json.dumps(info)
    }
    res = requests.post(
        JESSE_API_URL + '/exception',
        data=params,
        files=files,
        headers={'Authorization': f'Bearer {access_token}'}
    )

    success_message = 'Exception report submitted successfully'
    error_message = f"{res.status_code} error: {res.json()['message']}"

    return JSONResponse({
        'status': 'success' if res.status_code == 200 else 'error',
        'message': success_message if res.status_code == 200 else error_message
    }, status_code=200)



from typing import Union

import numpy as np

from jesse.helpers import get_candle_source, same_length, slice_candles
import jesse_rust


def zlema(candles: np.ndarray, period: int = 20, source_type: str = "close", sequential: bool = False) -> Union[
        float, np.ndarray]:
    """
    Zero-Lag Exponential Moving Average

    :param candles: np.ndarray
    :param period: int - default: 20
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    if len(candles.shape) == 1:
        source = candles
    else:
        candles = slice_candles(candles, sequential)
        source = get_candle_source(candles, source_type=source_type)
        
    # Use the Rust implementation
    res = jesse_rust.zlema(source, period)

    return same_length(candles, res) if sequential else res[-1]



from jesse.strategies import Strategy


# test_on_route_increased_position_and_on_route_reduced_position_and_strategy_vars part 1 - BTC-USD
class Test29(Strategy):
    def __init__(self) -> None:
        super().__init__()

        self.vars['should_short'] = False
        self.vars['should_long'] = False

    def should_long(self):
        return self.vars['should_long']

    def should_short(self):
        return self.vars['should_short']

    def go_long(self):
        self.buy = 1, self.price
        self.take_profit = 1, self.price + 10

    def go_short(self):
        self.sell = 1, self.price
        self.stop_loss = 1, self.price + 10

    def on_route_increased_position(self, strategy):
        # setting it to True means we'll open a position on NEXT candle
        self.vars['should_long'] = True

    def on_route_reduced_position(self, strategy):
        # setting it to True means we'll open a position on NEXT candle
        self.vars['should_short'] = True

    def should_cancel_entry(self):
        return False

    def on_close_position(self, order):
        self.vars['should_long'] = False
        self.vars['should_short'] = False



from typing import Union

import numpy as np

from jesse.helpers import get_candle_source, slice_candles


def roc(candles: np.ndarray, period: int = 10, source_type: str = "close", sequential: bool = False) -> Union[
    float, np.ndarray]:
    """
    ROC - Rate of change : ((price/prevPrice)-1)*100

    :param candles: np.ndarray
    :param period: int - default: 10
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    if len(candles.shape) == 1:
        source = candles
    else:
        candles = slice_candles(candles, sequential)
        source = get_candle_source(candles, source_type=source_type)

    n = len(source)
    res = np.full(n, np.nan, dtype=float)
    if n > period:
        res[period:] = (source[period:] / source[:-period] - 1) * 100

    return res if sequential else res[-1]



import numpy as np
from typing import Union

from jesse.helpers import get_candle_source, slice_candles
from jesse_rust import rsi as rsi_rust


def rsi(candles: np.ndarray, period: int = 14, source_type: str = "close", sequential: bool = False) -> Union[float, np.ndarray]:
    """
    RSI - Relative Strength Index

    :param candles: np.ndarray
    :param period: int - default: 14
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    if len(candles.shape) == 1:
        source = candles
    else:
        candles = slice_candles(candles, sequential)
        source = get_candle_source(candles, source_type=source_type)

    p = np.asarray(source, dtype=np.float64)
    result = rsi_rust(p, period)
    return result if sequential else result[-1]



from typing import Union

import numpy as np
from numba import njit

from jesse.helpers import slice_candles

@njit(cache=True)
def calculate_cci_loop(tp, period):
    n = tp.shape[0]
    result = np.empty(n)
    # initialize result with NaNs
    for i in range(n):
        result[i] = np.nan
    if n < period:
        return result
    for i in range(period - 1, n):
        sum_tp = 0.0
        for j in range(i - period + 1, i + 1):
            sum_tp += tp[j]
        sma = sum_tp / period
        sum_diff = 0.0
        for j in range(i - period + 1, i + 1):
            # Calculate absolute deviation
            if tp[j] >= sma:
                sum_diff += tp[j] - sma
            else:
                sum_diff += sma - tp[j]
        md = sum_diff / period
        if md == 0.0:
            result[i] = 0.0
        else:
            result[i] = (tp[i] - sma) / (0.015 * md)
    return result


def cci(candles: np.ndarray, period: int = 14, sequential: bool = False) -> Union[float, np.ndarray]:
    """
    CCI - Commodity Channel Index

    :param candles: np.ndarray
    :param period: int - default: 14
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)

    high = candles[:, 3]
    low = candles[:, 4]
    close = candles[:, 2]
    tp = (high + low + close) / 3.0

    result = calculate_cci_loop(tp, period)
    return result if sequential else result[-1]



from collections import namedtuple

import numpy as np
from numba import njit

from jesse.helpers import get_candle_source, slice_candles

ITREND = namedtuple('ITREND', ['signal', 'it', 'trigger'])


def itrend(candles: np.ndarray, alpha: float = 0.07, source_type: str = "hl2", sequential: bool = False) -> ITREND:
    """
    Instantaneous Trendline

    :param candles: np.ndarray
    :param alpha: float - default: 0.07
    :param source_type: str - default: "hl2"
    :param sequential: bool - default: False

    :return: ITREND(signal, it, trigger)
    """
    candles = slice_candles(candles, sequential)

    source = get_candle_source(candles, source_type=source_type)

    signal, it, trigger = itrend_fast(source, alpha)

    if sequential:
        return ITREND(signal, it, trigger)
    else:
        return ITREND(signal[-1], it[-1], trigger[-1])


@njit(cache=True)
def itrend_fast(source, alpha):
    it = np.copy(source)
    for i in range(2, 7):
        it[i] = (source[i] + 2 * source[i - 1] + source[i - 2]) / 4
    for i in range(7, source.shape[0]):
        it[i] = (alpha - alpha ** 2 / 4) * source[i] \
                + alpha ** 2 / 2 * source[i - 1] \
                - (alpha - alpha ** 2 * 3 / 4) * source[i - 2] \
                + 2 * (1 - alpha) * it[i - 1] - (1 - alpha) ** 2 * it[i - 2]

    # compute lead 2 trigger & signal
    lag2 = np.roll(it, 20)
    lag2[:20] = it[:20]
    trigger = 2 * it - lag2
    signal = (trigger > it) * 1 - (trigger < it) * 1
    return signal, it, trigger



from typing import Optional
from fastapi import APIRouter, Header
from fastapi.responses import JSONResponse

from jesse.services import auth as authenticator
from jesse.services.web import (
    NewStrategyRequestJson,
    GetStrategyRequestJson,
    SaveStrategyRequestJson,
    DeleteStrategyRequestJson
)
import jesse.helpers as jh

router = APIRouter(prefix="/strategy", tags=["Strategy"])


@router.post("/make")
def make_strategy(json_request: NewStrategyRequestJson, authorization: Optional[str] = Header(None)) -> JSONResponse:
    """
    Create a new strategy
    """
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    from jesse.services import strategy_handler
    return strategy_handler.generate(json_request.name)


@router.get("/all")
def get_strategies(authorization: Optional[str] = Header(None)) -> JSONResponse:
    """
    Get all strategies
    """
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    from jesse.services import strategy_handler
    return strategy_handler.get_strategies()


@router.post("/get")
def get_strategy(
        json_request: GetStrategyRequestJson,
        authorization: Optional[str] = Header(None)
) -> JSONResponse:
    """
    Get a specific strategy
    """
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    from jesse.services import strategy_handler
    return strategy_handler.get_strategy(json_request.name)


@router.post("/save")
def save_strategy(
        json_request: SaveStrategyRequestJson,
        authorization: Optional[str] = Header(None)
) -> JSONResponse:
    """
    Save a strategy
    """
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    from jesse.services import strategy_handler
    return strategy_handler.save_strategy(json_request.name, json_request.content)


@router.post("/delete")
def delete_strategy(
        json_request: DeleteStrategyRequestJson,
        authorization: Optional[str] = Header(None)
) -> JSONResponse:
    """
    Delete a strategy
    """
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    from jesse.services import strategy_handler
    return strategy_handler.delete_strategy(json_request.name)



from collections import namedtuple

import numpy as np
import jesse_rust
from jesse.helpers import slice_candles

VI = namedtuple('VI', ['plus', 'minus'])


def vi(candles: np.ndarray, period: int = 14, sequential: bool = False) -> VI:
    """
    Vortex Indicator (VI)

    :param candles: np.ndarray
    :param period: int - default: 14
    :param sequential: bool - default: False

    :return: VI(plus, minus)
    """
    candles = slice_candles(candles, sequential)
    
    # Use the Rust implementation
    vi_plus, vi_minus = jesse_rust.vi(candles, period, sequential)
    
    if sequential:
        return VI(vi_plus, vi_minus)
    else:
        return VI(vi_plus[-1], vi_minus[-1])



from typing import Union

import numpy as np
from numba import njit

from jesse.helpers import get_candle_source, slice_candles


def mwdx(candles: np.ndarray, factor: float = 0.2, source_type: str = "close", sequential: bool = False) -> Union[
    float, np.ndarray]:
    """
    MWDX Average

    :param candles: np.ndarray
    :param factor: float - default: 0.2
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """

    # Accept normal array too.
    if len(candles.shape) == 1:
        source = candles
    else:
        candles = slice_candles(candles, sequential)
        source = get_candle_source(candles, source_type=source_type)

    val2 = (2 / factor) - 1
    fac = 2 / (val2 + 1)

    res = mwdx_fast(source, fac)

    return res if sequential else res[-1]


@njit(cache=True)
def mwdx_fast(source, fac):
    newseries = np.copy(source)
    for i in range(1, source.shape[0]):
        newseries[i] = (fac * source[i]) + ((1 - fac) * newseries[i - 1])
    return newseries



from .ApexProMain import ApexProMain
from jesse.enums import exchanges


class ApexOmniPerpetualTestnet(ApexProMain):
    def __init__(self) -> None:
        super().__init__(
            name=exchanges.APEX_OMNI_PERPETUAL_TESTNET,
            rest_endpoint='https://testnet.omni.apex.exchange/api/v3'
        )



from jesse.testing_utils import single_route_backtest
import pytest
from jesse import exceptions
import jesse.helpers as jh


def test_should_be_able_to_short_in_spot_mode():
    with pytest.raises(exceptions.InvalidStrategy):
        single_route_backtest('TestShortInSpot', is_futures_trading=False)


def test_should_raise_exception_if_trying_to_spend_more_than_available_balance_in_spot_mode():
    with pytest.raises(exceptions.InsufficientBalance):
        single_route_backtest('TestCannotSpendMoreThanAvailableBalance', is_futures_trading=False)


def test_should_raise_exception_if_trying_to_submit_take_profit_order_with_size_more_than_current_position_qty():
    """
    cannot submit take-profit order with the size more than the current position's qty
    """
    with pytest.raises(exceptions.InsufficientBalance):
        single_route_backtest(
            'TestCannotSubmitTakeProfitOrderWithSizeMoreThanCurrentPositionQty',
            is_futures_trading=False
        )


def test_should_raise_exception_if_trying_to_submit_stop_loss_order_with_size_more_than_current_position_qty():
    """
    cannot submit stop-loss order with the size more than the current position's qty
    """
    with pytest.raises(exceptions.InsufficientBalance):
        single_route_backtest(
            'TestCannotSubmitStopLossOrderWithSizeMoreThanCurrentPositionQty',
            is_futures_trading=False
        )


def test_should_be_able_to_submit_take_profit_order_with_size_less_or_equal_to_current_position_qty():
    """
    test that can indeed submit a take profit order with size less or equal to the current position's qty
    """
    single_route_backtest(
        'TestCanSubmitTakeProfitOrderWithSizeEqualToCurrentPositionQty',
        is_futures_trading=False
    )

    single_route_backtest(
        'TestCanSubmitTakeProfitOrderWithSizeLessThanCurrentPositionQty',
        is_futures_trading=False
    )


def test_should_be_able_to_submit_stop_loss_order_with_size_less_or_equal_to_current_position_qty():
    """
    test that can indeed submit a stop-loss order with size less or equal to the current position's qty
    """
    single_route_backtest(
        'TestCanSubmitStopLossOrderWithSizeEqualToCurrentPositionQty',
        is_futures_trading=False,
        trend='down'
    )

    single_route_backtest(
        'TestCanSubmitStopLossOrderWithSizeLessThanCurrentPositionQty',
        is_futures_trading=False,
        trend='down'
    )


def test_should_be_able_submit_take_profit_and_stop_loss_at_same_time_in_spot():
    single_route_backtest(
        'TestCanSubmitTakeProfitAndStopLossAtSameTimeInSpot',
        is_futures_trading=False
    )


def test_stop_order_should_consider_executed_take_profit_orders_in_spot():
    """
    test that stop-loss order considers executed take-profit orders (when
    the take-profit order is submitted via multiple orders.
    """
    single_route_backtest(
        'TestStopOrderShouldConsiderExecutedTakeProfitOrdersInSpot',
        is_futures_trading=False
    )


def test_balances_are_handled_correctly_for_cancelling_orders_in_spot():
    single_route_backtest(
        'TestBalancesAreHandledCorrectlyForCancellingOrdersInSpot',
        is_futures_trading=False
    )


def test_position_open_increase_reduce_close_events_in_spot():
    single_route_backtest(
        'TestPositionOpenIncreaseReduceCloseEventsInSpot',
        is_futures_trading=False,
        fee=0.001
    )


def test_balance_and_fee_reduction_works_correctly_in_spot_mode_in_both_buy_and_sell_orders():
    single_route_backtest(
        'TestBalanceAndFeeReductionWorksCorrectlyInSpotModeInBothBuyAndSellOrders',
        is_futures_trading=False,
        fee=0.001
    )


def test_cannot_set_stop_loss_order_in_go_long():
    with pytest.raises(exceptions.InvalidStrategy):
        single_route_backtest(
            'TestCannotSetStopLossOrderInGoLong',
            is_futures_trading=False
        )


def test_cannot_set_take_profit_order_in_go_long():
    with pytest.raises(exceptions.InvalidStrategy):
        single_route_backtest(
            'TestCannotSetTakeProfitOrderInGoLong',
            is_futures_trading=False
        )



class Route:
    def __init__(
            self,
            exchange: str,
            symbol: str,
            timeframe: str = None,
            strategy_name: str = None,
            dna: str = None
    ) -> None:
        self.exchange = exchange
        self.symbol = symbol
        self.timeframe = timeframe
        self.strategy_name = strategy_name
        self.strategy = None
        self.dna = dna



from typing import Union

import numpy as np

from jesse.helpers import get_candle_source, slice_candles


def ui(candles: np.ndarray, period: int = 14, scalar: float = 100, source_type: str = "close",  sequential: bool = False) -> Union[float, np.ndarray]:
    """
    Ulcer Index (UI)

    :param candles: np.ndarray
    :param period: int - default: 14
    :param scalar: float - default: 100
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)
    source = get_candle_source(candles, source_type=source_type)
    n = source.shape[0]

    # Compute rolling maximum over the period
    if n < period:
        highest_close = np.full_like(source, np.nan)
    else:
        # sliding_window_view creates a rolling window view: shape (n-period+1, period)
        highest_window = np.lib.stride_tricks.sliding_window_view(source, window_shape=period)
        highest_max = np.max(highest_window, axis=1)
        highest_close = np.concatenate((np.full(period - 1, np.nan), highest_max))

    # Calculate downside percentage
    downside = scalar * (source - highest_close) / highest_close
    d2 = downside ** 2

    # Compute rolling sum of squared downside values
    if n < period:
        rolling_d2_sum = np.full_like(d2, np.nan)
    else:
        rolling_sum_valid = np.convolve(d2, np.ones(period), mode='valid')
        rolling_d2_sum = np.concatenate((np.full(period - 1, np.nan), rolling_sum_valid))

    res = np.sqrt(rolling_d2_sum / period)

    return res if sequential else res[-1]



from collections import namedtuple

import numpy as np
from numpy.lib.stride_tricks import sliding_window_view

from jesse.helpers import np_shift, slice_candles

IchimokuCloud = namedtuple('IchimokuCloud',
                           ['conversion_line', 'base_line', 'span_a', 'span_b', 'lagging_line', 'future_span_a',
                            'future_span_b'])


def ichimoku_cloud_seq(candles: np.ndarray, conversion_line_period: int = 9, base_line_period: int = 26,
                       lagging_line_period: int = 52, displacement: int = 26,
                       sequential: bool = False) -> IchimokuCloud:
    """
    Ichimoku Cloud

    :param candles: np.ndarray
    :param conversion_line_period: int - default: 9
    :param base_line_period: int - default: 26
    :param lagging_line_period: int - default: 52
    :param displacement: - default: 26
    :param sequential: bool - default: False

    :return: IchimokuCloud
    """

    if candles.shape[0] < lagging_line_period + displacement:
        raise ValueError("Too few candles available for lagging_line_period + displacement.")

    candles = slice_candles(candles, sequential)

    conversion_line = _line_helper(candles, conversion_line_period)
    base_line = _line_helper(candles, base_line_period)
    span_b_pre = _line_helper(candles, lagging_line_period)
    span_b = np_shift(span_b_pre, displacement, fill_value=np.nan)
    span_a_pre = (conversion_line + base_line) / 2
    span_a = np_shift(span_a_pre, displacement, fill_value=np.nan)
    lagging_line = np_shift(candles[:, 2], displacement - 1, fill_value=np.nan)

    if sequential:
        return IchimokuCloud(conversion_line, base_line, span_a, span_b, lagging_line, span_a_pre, span_b_pre)
    else:
        return IchimokuCloud(conversion_line[-1], base_line[-1], span_a[-1], span_b[-1], lagging_line[-1],
                             span_a_pre[-1], span_b_pre[-1])

def _line_helper(candles, period):
    small_ph = _rolling_max(candles[:, 3], period)
    small_pl = _rolling_min(candles[:, 4], period)
    return (small_ph + small_pl) / 2

def _rolling_max(a, period):
    n = len(a)
    if n < period:
        return np.full(n, np.nan)
    windows = sliding_window_view(a, window_shape=period)
    r = np.empty(n, dtype=a.dtype)
    r[:period-1] = np.nan
    r[period-1:] = np.max(windows, axis=-1)
    return r

def _rolling_min(a, period):
    n = len(a)
    if n < period:
        return np.full(n, np.nan)
    windows = sliding_window_view(a, window_shape=period)
    r = np.empty(n, dtype=a.dtype)
    r[:period-1] = np.nan
    r[period-1:] = np.min(windows, axis=-1)
    return r



import jesse.helpers as jh
from jesse.modes.utils import get_exchange_type
from jesse.enums import exchanges
from jesse.info import exchange_info

# Main configuration used by the Jesse framework. These values are modified
# at runtime based on the mode (backtest, live, or optimize) and user settings.
config = {
    # these values are related to the user's environment
    'env': {
        'caching': {
            'driver': 'pickle'
        },

        'logging': {
            'strategy_execution': True,
            'order_submission': True,
            'order_cancellation': True,
            'order_execution': True,
            'position_opened': True,
            'position_increased': True,
            'position_reduced': True,
            'position_closed': True,
            'shorter_period_candles': False,
            'trading_candles': True,
            'balance_update': True,
            'exchange_ws_reconnection': True
        },

        # fill it later in this file using data in info.py
        'exchanges': {
            exchanges.SANDBOX: {
                'fee': 0,
                'type': 'futures',
                # accepted values are: 'cross' and 'isolated'
                'futures_leverage_mode': 'cross',
                # 1x, 2x, 10x, 50x, etc. Enter as integers
                'futures_leverage': 1,
                'balance': 10_000,
            },
        },

        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
        # Optimize mode (using Optuna)
        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
        #
        # Below configurations are related to the optimize mode
        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
        'optimization': {
            # available ratio options: sharpe, calmar, sortino, omega, serenity, smart sharpe, smart sortino
            'objective_function': 'sharpe',
            # number of trials per each hyperparameter
            'trials': 200,
        },

        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
        # Data
        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
        #
        # Below configurations are related to the data
        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
        'data': {
            # The minimum number of warmup candles that is loaded before each session.
            'warmup_candles_num': 240,
            'generate_candles_from_1m': False,
            'persistency': True,
        },
    },

    # These values are just placeholders used by Jesse at runtime
    'app': {
        # list of currencies to consider
        'considering_symbols': [],
        # The symbol to trade.
        'trading_symbols': [],

        # list of time frames to consider
        'considering_timeframes': [],
        # Which candle type do you intend trade on
        'trading_timeframes': [],

        # list of exchanges to consider
        'considering_exchanges': [],
        # list of exchanges to consider
        'trading_exchanges': [],

        'considering_candles': [],

        # dict of registered live trade drivers
        'live_drivers': {},

        # Accepted values are: 'backtest', 'livetrade', 'fitness'.
        'trading_mode': '',

        # this would enable many console.log()s in the code, which are helpful for debugging.
        'debug_mode': False,

        # this is only used for the live unit tests
        'is_unit_testing': False,
    },
}

# set exchange config values based on the info
for key in exchange_info:
    config['env']['exchanges'][key] = {
        'fee': exchange_info[key]['fee'],
        'type': exchange_info[key]['type'],
        'futures_leverage_mode': 'cross',
        'futures_leverage': 1,
        'balance': 10_000
    }


def set_config(conf: dict) -> None:
    global config

    # optimization mode only
    if jh.is_optimizing():
        # objective function
        if 'objective_function' in conf:
            config['env']['optimization']['objective_function'] = conf['objective_function']
        # warm_up_candles
        config['env']['data']['warmup_candles_num'] = int(conf['warm_up_candles'])
        # number of trials per each hyperparameter
        config['env']['optimization']['trials'] = int(conf['trials'])

    # backtest and live
    if jh.is_backtesting() or jh.is_live():
        # warm_up_candles
        config['env']['data']['warmup_candles_num'] = int(conf['warm_up_candles'])
        # logs
        config['env']['logging'] = conf['logging']
        # exchanges
        for key, e in conf['exchanges'].items():
            if not jh.is_live() and e['type']:
                exchange_type = e['type']
            else:
                exchange_type = get_exchange_type(e['name'])
            config['env']['exchanges'][e['name']] = {
                'fee': float(e['fee']),
                'type': exchange_type,
                'balance': float(e['balance'])
            }
            if config['env']['exchanges'][e['name']]['type'] == 'futures':
                # 1x, 2x, 10x, 50x, etc. Enter as integers
                config['env']['exchanges'][e['name']]['futures_leverage'] = int(e.get('futures_leverage', 1))
                # accepted values are: 'cross' and 'isolated'
                config['env']['exchanges'][e['name']]['futures_leverage_mode'] = e.get('futures_leverage_mode', 'cross')

    # live mode only
    if jh.is_live():
        config['env']['notifications'] = conf['notifications']
        config['env']['data']['persistency'] = conf['persistency']
        config['env']['data']['generate_candles_from_1m'] = conf['generate_candles_from_1m']

    # TODO: must become a config value later when we go after multi account support?
    config['env']['identifier'] = 'main'


def reset_config() -> None:
    global config
    config = backup_config.copy()


backup_config = config.copy()



from .ApexProMain import ApexProMain
from jesse.enums import exchanges


class ApexOmniPerpetual(ApexProMain):
    def __init__(self) -> None:
        super().__init__(
            name=exchanges.APEX_OMNI_PERPETUAL,
            rest_endpoint='https://omni.apex.exchange/api/v3'
        )



from jesse.strategies import Strategy


# test_validation_for_equal_stop_loss_and_take_profit
class Test46(Strategy):
    def should_long(self) -> bool:
        return self.index == 0

    def should_short(self) -> bool:
        return False

    def go_long(self):
        self.buy = 1, 2

    def update_position(self):
        if self.index == 5:
            self.stop_loss = 1, 3
            self.take_profit = 1, 3

    def go_short(self):
        pass

    def should_cancel_entry(self):
        return False



from typing import Optional
from fastapi import APIRouter, Header, Query
from fastapi.responses import JSONResponse, FileResponse

from jesse.services import auth as authenticator
from jesse.services.multiprocessing import process_manager
from jesse.services.web import BacktestRequestJson, CancelRequestJson
import jesse.helpers as jh

router = APIRouter(prefix="/backtest", tags=["Backtest"])


@router.post("")
def backtest(request_json: BacktestRequestJson, authorization: Optional[str] = Header(None)):
    """
    Start a backtest process
    """
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    jh.validate_cwd()

    from jesse.modes.backtest_mode import run as run_backtest

    process_manager.add_task(
        run_backtest,
        request_json.id,
        request_json.debug_mode,
        request_json.config,
        request_json.exchange,
        request_json.routes,
        request_json.data_routes,
        request_json.start_date,
        request_json.finish_date,
        None,
        request_json.export_chart,
        request_json.export_tradingview,
        request_json.export_csv,
        request_json.export_json,
        request_json.fast_mode,
        request_json.benchmark
    )

    return JSONResponse({'message': 'Started backtesting...'}, status_code=202)


@router.post("/cancel")
def cancel_backtest(request_json: CancelRequestJson, authorization: Optional[str] = Header(None)):
    """
    Cancel a backtest process
    """
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    process_manager.cancel_process(request_json.id)

    return JSONResponse({'message': f'Backtest process with ID of {request_json.id} was requested for termination'},
                        status_code=202)


@router.get("/logs/{session_id}")
def get_logs(session_id: str, token: str = Query(...)):
    """
    Get logs as text for a specific session. Similar to download but returns text content instead of file.
    """
    if not authenticator.is_valid_token(token):
        return authenticator.unauthorized_response()

    try:
        from jesse.modes.data_provider import get_backtest_logs
        content = get_backtest_logs(session_id)

        if content is None:
            return JSONResponse({'error': 'Log file not found'}, status_code=404)

        return JSONResponse({'content': content}, status_code=200)
    except Exception as e:
        return JSONResponse({'error': str(e)}, status_code=500)


@router.get("/download-log/{session_id}")
def download_backtest_log(session_id: str, token: str = Query(...)):
    """
    Download log file for a specific backtest session
    """
    if not authenticator.is_valid_token(token):
        return authenticator.unauthorized_response()

    try:
        from jesse.modes.data_provider import download_backtest_log
        return download_backtest_log(session_id)
    except Exception as e:
        return JSONResponse({'error': str(e)}, status_code=500)



@router.get("/logs/{session_id}")
def get_logs(session_id: str, token: str = Query(...)):
    """
    Get logs as text for a specific session. Similar to download but returns text content instead of file.
    """
    if not authenticator.is_valid_token(token):
        return authenticator.unauthorized_response()

    try:
        from jesse.modes.data_provider import get_backtest_logs
        content = get_backtest_logs(session_id)

        if content is None:
            return JSONResponse({'error': 'Log file not found'}, status_code=404)

        return JSONResponse({'content': content}, status_code=200)
    except Exception as e:
        return JSONResponse({'error': str(e)}, status_code=500)




from collections import namedtuple

import numpy as np

from jesse.helpers import slice_candles
from jesse.indicators.ma import ma

KDJ = namedtuple('KDJ', ['k', 'd', 'j'])

def _rolling_max(a, window):
    from numpy.lib.stride_tricks import sliding_window_view
    a = np.asarray(a)
    if len(a) < window:
        return np.maximum.accumulate(a)
    result = np.empty_like(a)
    # Use vectorized cumulative maximum for the first window-1 elements
    result[:window-1] = np.maximum.accumulate(a)[:window-1]
    windows = sliding_window_view(a, window_shape=window)
    result[window-1:] = np.max(windows, axis=1)
    return result

def _rolling_min(a, window):
    from numpy.lib.stride_tricks import sliding_window_view
    a = np.asarray(a)
    if len(a) < window:
        return np.minimum.accumulate(a)
    result = np.empty_like(a)
    # Use vectorized cumulative minimum for the first window-1 elements
    result[:window-1] = np.minimum.accumulate(a)[:window-1]
    windows = sliding_window_view(a, window_shape=window)
    result[window-1:] = np.min(windows, axis=1)
    return result

def kdj(candles: np.ndarray, fastk_period: int = 9, slowk_period: int = 3, slowk_matype: int = 0,
          slowd_period: int = 3, slowd_matype: int = 0, sequential: bool = False) -> KDJ:
    """
    The KDJ Oscillator

    :param candles: np.ndarray
    :param fastk_period: int - default: 9
    :param slowk_period: int - default: 3
    :param slowk_matype: int - default: 0
    :param slowd_period: int - default: 3
    :param slowd_matype: int - default: 0
    :param sequential: bool - default: False

    :return: KDJ(k, d, j)
    """
    if any(matype in (24, 29) for matype in (slowk_matype, slowd_matype)):
        raise ValueError("VWMA (matype 24) and VWAP (matype 29) cannot be used in kdj indicator.")
    
    candles = slice_candles(candles, sequential)

    candles_close = candles[:, 2]
    candles_high = candles[:, 3]
    candles_low = candles[:, 4]

    hh = _rolling_max(candles_high, fastk_period)
    ll = _rolling_min(candles_low, fastk_period)

    stoch = 100 * (candles_close - ll) / (hh - ll)
    
    k = ma(stoch, period=slowk_period, matype=slowk_matype, sequential=True)
    d = ma(k, period=slowd_period, matype=slowd_matype, sequential=True)
    j = 3 * k - 2 * d

    if sequential:
        return KDJ(k, d, j)
    else:
        return KDJ(k[-1], d[-1], j[-1])



from collections import namedtuple

import numpy as np

from jesse.helpers import get_candle_source, np_shift, slice_candles

GATOR = namedtuple('GATOR', ['upper', 'lower', 'upper_change', 'lower_change'])


def gatorosc(candles: np.ndarray, source_type: str = "close", sequential: bool = False) -> GATOR:
    """
    Gator Oscillator by Bill M. Williams

    :param candles: np.ndarray
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: GATOR(upper, lower, upper_change, lower_change)
    """

    candles = slice_candles(candles, sequential)

    source = get_candle_source(candles, source_type=source_type)

    jaw = np_shift(numpy_ewma(source, 13), 8, fill_value=np.nan)
    teeth = np_shift(numpy_ewma(source, 8), 5, fill_value=np.nan)
    lips = np_shift(numpy_ewma(source, 5), 3, fill_value=np.nan)

    upper = np.abs(jaw - teeth)
    lower = -np.abs(teeth - lips)

    # Calculate momentum for upper: difference from previous value, first element is np.nan
    upper_change = np.empty_like(upper)
    upper_change[0] = np.nan
    upper_change[1:] = np.diff(upper)

    # Calculate momentum for lower: negative difference from previous value, first element is np.nan
    lower_change = np.empty_like(lower)
    lower_change[0] = np.nan
    lower_change[1:] = -np.diff(lower)

    if sequential:
        return GATOR(upper, lower, upper_change, lower_change)
    else:
        return GATOR(upper[-1], lower[-1], upper_change[-1], lower_change[-1])


def numpy_ewma(data, window):
    """

    :param data:
    :param window:
    :return:
    """
    alpha = 1 / window
    # scale = 1 / (1 - alpha)
    n = data.shape[0]
    scale_arr = (1 - alpha) ** (-1 * np.arange(n))
    weights = (1 - alpha) ** np.arange(n)
    pw0 = (1 - alpha) ** (n - 1)
    mult = data * pw0 * scale_arr
    cumsums = mult.cumsum()
    return cumsums * scale_arr[::-1] / weights.cumsum()



from hashlib import sha256
from fastapi.responses import JSONResponse
from jesse.services.env import ENV_VALUES


def password_to_token(password: str) -> JSONResponse:
    if password != ENV_VALUES['PASSWORD']:
        return unauthorized_response()

    auth_token = sha256(password.encode('utf-8')).hexdigest()

    return JSONResponse({
        'auth_token': auth_token,
    }, status_code=200)


def is_valid_token(auth_token: str) -> bool:
    hashed_local_pass = sha256(ENV_VALUES['PASSWORD'].encode('utf-8')).hexdigest()
    return auth_token == hashed_local_pass


def unauthorized_response() -> JSONResponse:
    return JSONResponse({
        'message': "Invalid password",
    }, status_code=401)


def get_access_token():
    from jesse.services.env import ENV_VALUES

    if 'LICENSE_API_TOKEN' not in ENV_VALUES:
        return None
    if not ENV_VALUES['LICENSE_API_TOKEN']:
        return None

    return ENV_VALUES['LICENSE_API_TOKEN']




import asyncio
import json
from typing import Set
from starlette.websockets import WebSocket

from jesse.services.redis import async_redis
from jesse.services.multiprocessing import process_manager
import jesse.helpers as jh


class ConnectionManager:
    def __init__(self):
        self.active_connections: Set[WebSocket] = set()
        self.is_subscribed = False
        self.redis_subscriber = None
        self.reader_task = None
        
    async def connect(self, websocket: WebSocket):
        await websocket.accept()
        self.active_connections.add(websocket)
        
    def disconnect(self, websocket: WebSocket):
        self.active_connections.remove(websocket)
        
    async def broadcast(self, message: dict):
        # Process id with the manager for each websocket separately
        for connection in self.active_connections:
            message_copy = dict(message)
            message_copy['id'] = process_manager.get_client_id(message_copy['id'])
            await connection.send_json(message_copy)
            
    async def start_redis_listener(self, channel_pattern):
        if not self.is_subscribed:
            self.redis_subscriber, = await async_redis.psubscribe(channel_pattern)
            self.is_subscribed = True
            
            # Start the listener task
            self.reader_task = asyncio.create_task(self._redis_listener(self.redis_subscriber))
            
    async def _redis_listener(self, channel):
        try:
            async for ch, message in channel.iter():
                # Parse the message and broadcast to all clients
                message_dict = json.loads(message)
                await self.broadcast(message_dict)
        except Exception as e:
            print(jh.color(f"Redis listener error: {str(e)}", 'red'))
            
    async def stop_redis_listener(self):
        if self.is_subscribed and len(self.active_connections) == 0:
            # Only unsubscribe if no clients are connected
            if self.reader_task and not self.reader_task.done():
                self.reader_task.cancel()
                
            from jesse.services.env import ENV_VALUES
            await async_redis.punsubscribe(f"{ENV_VALUES['APP_PORT']}:channel:*")
            self.is_subscribed = False
            print(jh.color("Redis unsubscribed - no more active connections", 'yellow'))


# Create a global instance
ws_manager = ConnectionManager()



import peewee
import jesse.helpers as jh
import numpy as np


class Ticker(peewee.Model):
    id = peewee.UUIDField(primary_key=True)
    # timestamp in milliseconds
    timestamp = peewee.BigIntegerField()
    # the latest trades price
    last_price = peewee.FloatField()
    # the trading volume in the last 24 hours
    volume = peewee.FloatField()
    # the highest price in the last 24 hours
    high_price = peewee.FloatField()
    # the lowest price in the last 24 hours
    low_price = peewee.FloatField()
    symbol = peewee.CharField()
    exchange = peewee.CharField()

    class Meta:
        from jesse.services.db import database

        database = database.db
        indexes = ((('exchange', 'symbol', 'timestamp'), True),)

    def __init__(self, attributes: dict = None, **kwargs) -> None:
        peewee.Model.__init__(self, attributes=attributes, **kwargs)

        if attributes is None:
            attributes = {}

        for a, value in attributes.items():
            setattr(self, a, value)


# # # # # # # # # # # # # # # # # # # # # # # # # # # 
# # # # # # # # # DB FUNCTIONS # # # # # # # # #
# # # # # # # # # # # # # # # # # # # # # # # # # # # 

def store_ticker_into_db(exchange: str, symbol: str, ticker: np.ndarray) -> None:
    return
    d = {
        'id': jh.generate_unique_id(),
        'timestamp': ticker[0],
        'last_price': ticker[1],
        'high_price': ticker[2],
        'low_price': ticker[3],
        'volume': ticker[4],
        'symbol': symbol,
        'exchange': exchange,
    }

    def async_save() -> None:
        Ticker.insert(**d).on_conflict_ignore().execute()
        print(
            jh.color(f'ticker: {jh.timestamp_to_time(d["timestamp"])}-{exchange}-{symbol}: {ticker}', 'yellow')
        )

    # async call
    threading.Thread(target=async_save).start()



import numpy as np
from typing import Union

from jesse.helpers import slice_candles


def natr(candles: np.ndarray, period: int = 14, sequential: bool = False) -> Union[float, np.ndarray]:
    """
    NATR - Normalized Average True Range

    :param candles: np.ndarray
    :param period: int - default: 14
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)
    high = candles[:, 3]
    low = candles[:, 4]
    close = candles[:, 2]

    # Compute True Range (TR)
    n = len(candles)
    tr = np.empty(n, dtype=float)
    tr[0] = high[0] - low[0]
    if n > 1:
        diff1 = high[1:] - low[1:]
        diff2 = np.abs(high[1:] - close[:-1])
        diff3 = np.abs(low[1:] - close[:-1])
        tr[1:] = np.maximum(np.maximum(diff1, diff2), diff3)

    # Initialize ATR array
    atr = np.empty(n, dtype=float)
    atr[:period-1] = np.nan  # not enough data for smoothing
    base = np.mean(tr[:period])
    atr[period-1] = base

    # If there's no additional data after the initial period, return the current NATR
    if n == period:
        result = (base / close[period-1]) * 100
        return result if not sequential else np.concatenate((np.full(period-1, np.nan), [result]))

    # Wilder's smoothing is equivalent to an exponential moving average with alpha = 1/period
    alpha = 1.0 / period
    beta = 1 - alpha
    # For indices from period to end, we compute the recursive ATR as:
    # ATR[t] = beta^(t - (period-1)) * base + sum_{i=period}^{t} (alpha * beta^(t-i) * tr[i])
    # We vectorize the summation using convolution.
    x = tr[period:]
    m = len(x)
    weights = alpha * beta ** np.arange(m)
    conv = np.convolve(x, weights, mode='full')[:m]
    base_adjustment = beta ** (np.arange(1, m + 1)) * base
    atr[period:] = base_adjustment + conv

    natr_arr = (atr / close) * 100
    return natr_arr if sequential else natr_arr[-1]



from jesse.strategies import Strategy
import jesse.helpers as jh
from jesse import utils


class TestDailyBalanceStoresPortfolioValue(Strategy):
    def before(self) -> None:
        if self.index == 0:
            assert self.balance == 10_000

        if self.price == 11:
            assert self.portfolio_value == 10_000
            if self.is_spot_trading:
                assert self.balance == 10_000 - (50 * 9)

    def should_long(self) -> bool:
        return self.price == 10

    def go_long(self) -> None:
        # submit an entry buy order that is not supposed to be filled
        self.buy = 50, 9

    def should_cancel_entry(self):
        return False

    def before_terminate(self):
        # assert that all the items (daily balances) in store.app.daily_balance remain 10_000
        for item in self.daily_balances:
            assert item == 10_000



from typing import Union

import numpy as np
from jesse.helpers import slice_candles
from numba import njit


@njit(cache=True)
def _adxr(high, low, close, period):
    n = len(high)
    TR = np.zeros(n)
    DMP = np.zeros(n)
    DMM = np.zeros(n)
    
    # First value
    TR[0] = high[0] - low[0]
    
    # Calculate TR, DMP, DMM
    for i in range(1, n):
        hl = high[i] - low[i]
        hc = abs(high[i] - close[i-1])
        lc = abs(low[i] - close[i-1])
        TR[i] = max(hl, hc, lc)
        
        up_move = high[i] - high[i-1]
        down_move = low[i-1] - low[i]
        
        if up_move > down_move and up_move > 0:
            DMP[i] = up_move
        else:
            DMP[i] = 0
            
        if down_move > up_move and down_move > 0:
            DMM[i] = down_move
        else:
            DMM[i] = 0
    
    # Smoothed TR, DMP, DMM
    STR = np.zeros(n)
    S_DMP = np.zeros(n)
    S_DMM = np.zeros(n)
    
    # Initialize first value
    STR[0] = TR[0]
    S_DMP[0] = DMP[0]
    S_DMM[0] = DMM[0]
    
    # Calculate smoothed values
    for i in range(1, n):
        STR[i] = STR[i-1] - (STR[i-1] / period) + TR[i]
        S_DMP[i] = S_DMP[i-1] - (S_DMP[i-1] / period) + DMP[i]
        S_DMM[i] = S_DMM[i-1] - (S_DMM[i-1] / period) + DMM[i]
    
    # Calculate DI+ and DI-
    DI_plus = np.zeros(n)
    DI_minus = np.zeros(n)
    for i in range(n):
        if STR[i] != 0:
            DI_plus[i] = (S_DMP[i] / STR[i]) * 100
            DI_minus[i] = (S_DMM[i] / STR[i]) * 100
    
    # Calculate DX
    DX = np.zeros(n)
    for i in range(n):
        denom = DI_plus[i] + DI_minus[i]
        if denom != 0:
            DX[i] = (abs(DI_plus[i] - DI_minus[i]) / denom) * 100
    
    # Calculate ADX
    ADX = np.full(n, np.nan)
    if n >= period:
        for i in range(period-1, n):
            sum_dx = 0
            for j in range(period):
                sum_dx += DX[i-j]
            ADX[i] = sum_dx / period
    
    # Calculate ADXR
    ADXR = np.full(n, np.nan)
    if n > period:
        for i in range(period, n):
            ADXR[i] = (ADX[i] + ADX[i-period]) / 2
            
    return ADXR

def adxr(candles: np.ndarray, period: int = 14, sequential: bool = False) -> Union[float, np.ndarray]:
    """
    @author KivancOzbilgic
    credits: https://www.tradingview.com/v/9f5zDi3r/
    
    ADXR - Average Directional Movement Index Rating

    :param candles: np.ndarray with at least 5 columns where index 3 is high, index 4 is low and index 2 is close
    :param period: int - period for smoothing and moving average (default: 14)
    :param sequential: bool - returns full series if True, else only the last computed value
    :return: ADXR as float or np.ndarray
    """
    candles = slice_candles(candles, sequential)
    
    high = candles[:, 3]
    low = candles[:, 4]
    close = candles[:, 2]
    
    ADXR = _adxr(high, low, close, period)
    
    return ADXR if sequential else ADXR[-1]



from jesse.strategies import Strategy
from jesse.enums import order_types


class TestTakeProfitPriceIsReplacedWithMarketOrderWhenMoreConvenientShortPosition(Strategy):
    def before(self) -> None:
        if self.price == 15:
            last_trade = self.trades[-1]
            # it should have closed on the market price at the time being 10 instead of 12
            assert last_trade.exit_price == 10

            # the order type should be market
            assert self.trades[-1].orders[0].type == order_types.MARKET
            assert self.trades[-1].orders[1].type == order_types.MARKET

    def should_long(self) -> bool:
        return False

    def go_long(self):
        pass

    def should_short(self) -> bool:
        return self.price == 10

    def go_short(self):
        self.sell = 1, 10
        self.take_profit = 1, 12

    def should_cancel_entry(self):
        return False



from datetime import datetime, timedelta
from jesse.routes import router
from jesse.store import store
from jesse.services.candle import get_candles
from jesse.utils import prices_to_returns


def _calculate_equity_curve(daily_balance, start_date, name: str, color: str):
    date_list = [start_date + timedelta(days=x) for x in range(len(daily_balance))]
    eq = [{
        'time': date.timestamp(),
        'value': balance,
        'color': color
    } for date, balance in zip(date_list, daily_balance)]
    return {
        'name': name,
        'data': eq,
        'color': color,
    }


def _generate_color(previous_color):
    # Convert the previous color from hex to RGB
    previous_color = previous_color.lstrip('#')
    r, g, b = tuple(int(previous_color[i:i+2], 16) for i in (0, 2, 4))

    # Modify the RGB values to generate a new color
    r = (r + 50) % 256
    g = (g + 50) % 256
    b = (b + 50) % 256

    # Convert the new color from RGB to hex
    new_color = '#{:02x}{:02x}{:02x}'.format(r, g, b)

    return new_color


def equity_curve(benchmark: bool = False) -> list:
    if store.completed_trades.count == 0:
        return None

    result = []
    start_date = datetime.fromtimestamp(store.app.starting_time / 1000)
    daily_balance = store.app.daily_balance

    # Define the first 10 colors
    colors = ['#818CF8', '#fbbf24', '#fb7185', '#60A5FA', '#f472b6', '#A78BFA', '#f87171', '#6EE7B7', '#93C5FD', '#FCA5A5']

    result.append(_calculate_equity_curve(daily_balance, start_date, 'Portfolio', colors[0]))

    if benchmark:
        initial_balance = daily_balance[0]
        for i, r in enumerate(router.routes):
            _, daily_candles = get_candles(
                r.exchange, r.symbol, '1D', store.app.starting_time,
                store.app.ending_time + 1000 * 60 * 60 * 24, is_for_jesse=False, warmup_candles_num=0, caching=True
            )
            daily_returns = prices_to_returns(daily_candles[:, 2])
            daily_returns[0] = 0
            daily_balance_benchmark = initial_balance * (1 + daily_returns/100).cumprod()

            # If there are more than 10 routes, generate new colors
            if i + 1 >= 10:
                colors.append(_generate_color(colors[-1]))

            result.append(_calculate_equity_curve(daily_balance_benchmark, start_date, r.symbol, colors[(i + 1) % len(colors)]))

    return result



from collections import namedtuple

import numpy as np

from jesse.helpers import get_candle_source, slice_candles
from jesse.indicators.ma import ma

ERI = namedtuple('ERI', ['bull', 'bear'])


def eri(candles: np.ndarray, period: int = 13, matype: int = 1, source_type: str = "close",
        sequential: bool = False) -> ERI:
    """
    Elder Ray Index (ERI)

    :param candles: np.ndarray
    :param period: int - default: 13
    :param matype: int - default: 1
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)

    source = get_candle_source(candles, source_type=source_type)

    if matype == 24 or matype == 29:
        ema = ma(candles, period=period, matype=matype, source_type=source_type, sequential=True)
    else:
        ema = ma(source, period=period, matype=matype, sequential=True)

    bull = candles[:, 3] - ema
    bear = candles[:, 4] - ema

    if sequential:
        return ERI(bull, bear)
    else:
        return ERI(bull[-1], bear[-1])



from typing import Union

import numpy as np
from numba import njit

from jesse.helpers import same_length, slice_candles


def mass(candles: np.ndarray, period: int = 5, sequential: bool = False) -> Union[float, np.ndarray]:
    """
    MASS - Mass Index
    The Mass Index uses the high-low range to identify trend reversals based on range expansions.
    It suggests that a reversal of the current trend may be imminent when the range widens beyond 
    a certain point and then contracts.

    :param candles: np.ndarray
    :param period: int - default: 5
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)

    # Calculate high-low range
    high_low_range = candles[:, 3] - candles[:, 4]  # high - low

    # Calculate 9-period EMA of high-low range
    ema1 = calc_ema(high_low_range, 9)

    # Calculate 9-period EMA of the first EMA
    ema2 = calc_ema(ema1, 9)

    # Calculate EMA ratio
    ratio = np.divide(ema1, ema2, out=np.zeros_like(ema1), where=ema2 != 0)

    # Calculate period-sum of ratio using Numba for optimization
    res = mass_sum(ratio, period)

    return same_length(candles, res) if sequential else res[-1]


@njit(cache=True)
def mass_sum(ratio: np.ndarray, period: int) -> np.ndarray:
    """Calculate the sum of the ratio over the specified period"""
    result = np.zeros_like(ratio)
    for i in range(period - 1, len(ratio)):
        result[i] = np.sum(ratio[i-period+1:i+1])
    return result


@njit(cache=True)
def calc_ema(data, n):
    alpha = 2.0 / (n + 1)
    result = np.empty(data.shape[0])
    result[0] = data[0]
    for i in range(1, data.shape[0]):
        result[i] = alpha * data[i] + (1 - alpha) * result[i-1]
    return result




from collections import namedtuple

import numpy as np

from jesse.helpers import slice_candles

CKSP = namedtuple('CKSP', ['long', 'short'])

def atr(high: np.ndarray, low: np.ndarray, close: np.ndarray, timeperiod: int = 10) -> np.ndarray:
    tr = np.empty_like(close)
    tr[0] = high[0] - low[0]
    tr[1:] = np.maximum.reduce([
        high[1:] - low[1:],
        np.abs(high[1:] - close[:-1]),
        np.abs(low[1:] - close[:-1])
    ])
    atr_vals = np.empty_like(close)
    if len(close) < timeperiod:
        return np.full_like(close, np.nan)
    atr_vals[:timeperiod-1] = np.nan
    atr_vals[timeperiod-1] = np.mean(tr[:timeperiod])
    for t in range(timeperiod, len(close)):
        atr_vals[t] = (atr_vals[t-1]*(timeperiod-1) + tr[t]) / timeperiod
    return atr_vals

def rolling_max(arr: np.ndarray, window: int) -> np.ndarray:
    n = len(arr)
    if n == 0:
        return arr
    result = np.empty(n)
    if window > 1:
        result[:window-1] = np.maximum.accumulate(arr[:window-1])
        if n >= window:
            shape = (n - window + 1, window)
            strides = (arr.strides[0], arr.strides[0])
            windows = np.lib.stride_tricks.as_strided(arr, shape=shape, strides=strides)
            result[window-1:] = np.max(windows, axis=1)
    else:
        result = arr.copy()
    return result

def rolling_min(arr: np.ndarray, window: int) -> np.ndarray:
    n = len(arr)
    if n == 0:
        return arr
    result = np.empty(n)
    if window > 1:
        result[:window-1] = np.minimum.accumulate(arr[:window-1])
        if n >= window:
            shape = (n - window + 1, window)
            strides = (arr.strides[0], arr.strides[0])
            windows = np.lib.stride_tricks.as_strided(arr, shape=shape, strides=strides)
            result[window-1:] = np.min(windows, axis=1)
    else:
        result = arr.copy()
    return result

def cksp(candles: np.ndarray, p: int = 10, x: float = 1.0,  q: int = 9, sequential: bool = False) -> CKSP:
    """
    Chande Kroll Stop (CKSP)

    :param candles: np.ndarray
    :param p: int - default: 10 (ATR period)
    :param x: float - default: 1.0 (ATR multiplier)
    :param q: int - default: 9 (rolling window period)
    :param sequential: bool - default: False

    :return: CKSP namedtuple containing long and short values
    """
    candles = slice_candles(candles, sequential)

    candles_close = candles[:, 2]
    candles_high = candles[:, 3]
    candles_low = candles[:, 4]

    atr_vals = atr(candles_high, candles_low, candles_close, timeperiod=p)

    LS0 = rolling_max(candles_high, window=q) - x * atr_vals
    LS = rolling_max(LS0, window=q)

    SS0 = rolling_min(candles_low, window=q) + x * atr_vals
    SS = rolling_min(SS0, window=q)

    if sequential:
        return CKSP(LS, SS)
    else:
        return CKSP(LS[-1], SS[-1])




from typing import Union

import numpy as np

from jesse.helpers import slice_candles


def bop(candles: np.ndarray, sequential: bool = False) -> Union[float, np.ndarray]:
    """
    BOP - Balance Of Power

    :param candles: np.ndarray
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)

    open_prices = candles[:, 1]
    high_prices = candles[:, 3]
    low_prices = candles[:, 4]
    close_prices = candles[:, 2]
    denominator = high_prices - low_prices
    bop_values = np.where(denominator != 0, (close_prices - open_prices) / denominator, 0)

    return bop_values if sequential else bop_values[-1]



from typing import Union
import numpy as np
from jesse.helpers import get_candle_source, slice_candles
from jesse_rust import kama as kama_rust


def kama(candles: np.ndarray, period: int = 14, fast_length: int = 2, slow_length: int = 30, 
         source_type: str = "close", sequential: bool = False) -> Union[float, np.ndarray]:
    """
    KAMA - Kaufman Adaptive Moving Average
    
    :param candles: np.ndarray
    :param period: int - default: 14, lookback period for the calculation
    :param fast_length: int - default: 2, fast EMA length for smoothing factor
    :param slow_length: int - default: 30, slow EMA length for smoothing factor
    :param source_type: str - default: "close", specifies the candle field
    :param sequential: bool - default: False, if True returns the full array, otherwise only the last value

    :return: float | np.ndarray
    """
    if candles.ndim == 1:
        src = candles
    else:
        candles = slice_candles(candles, sequential)
        src = get_candle_source(candles, source_type=source_type)

    src = np.asarray(src, dtype=np.float64)
    
    result = kama_rust(src, period, fast_length, slow_length)
    
    return result if sequential else result[-1]



from jesse.strategies import Strategy
import jesse.helpers as jh
from jesse import utils


class TestCapitalPropertyRaisesNotImplementedError(Strategy):
    def should_long(self) -> bool:
        self.capital
        return False

    def go_long(self) -> None:
        pass

    def should_cancel_entry(self):
        return False



from jesse.strategies import Strategy


# test_filters
class Test37(Strategy):
    def before(self):
        """used it to do assertions"""
        if self.index in [3, 11]:
            assert self.take_profit is None
            assert self.stop_loss is None

    def should_long(self) -> bool:
        return self.index == 0

    def should_short(self) -> bool:
        return self.index == 10

    def go_long(self):
        qty = 1
        self.buy = qty, self.price
        self.stop_loss = qty, self.price - .10

    def go_short(self):
        qty = 1
        self.sell = qty, self.price
        self.take_profit = qty, self.price + .10

    def should_cancel_entry(self):
        return False

    def filters(self):
        return [
            self.filter_1
        ]

    def filter_1(self):
        if self.index == 0:
            return False

        if self.index == 10:
            return False

        return True



from typing import Union

import numpy as np
from numba import njit

from jesse.helpers import get_candle_source, same_length, slice_candles


@njit
def _wilders_fast(source: np.ndarray, period: int) -> np.ndarray:
    # Pre-allocate the output array
    res = np.zeros_like(source)
    # First value is a simple copy
    res[0] = source[0]

    # Calculate Wilder's Smoothing
    for i in range(1, len(source)):
        res[i] = (res[i - 1] * (period - 1) + source[i]) / period

    return res


def wilders(candles: np.ndarray, period: int = 5, source_type: str = "close", sequential: bool = False) -> Union[
        float, np.ndarray]:
    """
    WILDERS - Wilders Smoothing

    :param candles: np.ndarray
    :param period: int - default: 5
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    if len(candles.shape) == 1:
        source = candles
    else:
        candles = slice_candles(candles, sequential)
        source = get_candle_source(candles, source_type=source_type)

    res = _wilders_fast(source, period)

    return same_length(candles, res) if sequential else res[-1]



import jesse.services.selectors as selectors
from jesse.config import config, reset_config
from jesse.enums import exchanges
from jesse.store import store
from jesse.routes import router


def set_up():
    reset_config()
    config['app']['considering_exchanges'] = [exchanges.SANDBOX]
    config['app']['trading_exchanges'] = [exchanges.SANDBOX]
    config['env']['exchanges'][exchanges.SANDBOX]['balance'] = 2000
    routes = [
        {'exchange': exchanges.SANDBOX, 'symbol': 'BTC-USDT', 'timeframe': '1m', 'strategy': 'TestVanillaStrategy'}
    ]
    router.initiate(routes)


def test_have_correct_exchanges_in_store_after_creating_store():
    set_up()

    e = selectors.get_exchange(exchanges.SANDBOX)
    assert len(store.exchanges.storage) == 1
    assert e.assets['USDT'] == 2000



import numpy as np

import jesse.helpers as jh
from jesse.config import config, reset_config
from jesse.store import store


def set_up():
    """

    """
    reset_config()
    config['app']['considering_candles'] = [('Sandbox', 'BTC-USD')]
    store.reset()
    store.trades.init_storage()


def test_can_add_add_trade():
    set_up()

    np.testing.assert_equal(store.trades.get_trades('Sandbox', 'BTC-USD'), np.zeros((0, 6)))

    # add first trade
    t1 = np.array([jh.now_to_timestamp(), 100, 2, 1], dtype=np.float64)
    t2 = np.array([jh.now_to_timestamp(), 98, 2, 1], dtype=np.float64)
    t3 = np.array([jh.now_to_timestamp(), 98, 2, 0], dtype=np.float64)
    t4 = np.array([jh.now_to_timestamp(), 98, 2, 0], dtype=np.float64)
    t5 = np.array([jh.now_to_timestamp(), 98, 2, 0], dtype=np.float64)
    store.trades.add_trade(t1, 'Sandbox', 'BTC-USD')
    store.trades.add_trade(t2, 'Sandbox', 'BTC-USD')
    store.trades.add_trade(t3, 'Sandbox', 'BTC-USD')
    store.trades.add_trade(t4, 'Sandbox', 'BTC-USD')
    store.trades.add_trade(t5, 'Sandbox', 'BTC-USD')

    assert len(store.trades.get_trades('Sandbox', 'BTC-USD')) == 0

    t6 = np.array([jh.now_to_timestamp() + 1000, 98, 2, 1], dtype=np.float64)
    store.trades.add_trade(t6, 'Sandbox', 'BTC-USD')

    assert len(store.trades.get_trades('Sandbox', 'BTC-USD')) == 1

    np.testing.assert_equal(store.trades.get_current_trade('Sandbox', 'BTC-USD'), np.array([
        jh.now_to_timestamp(),
        # price
        (100 * 2 + 98 * 2 + 98 * 2 + 98 * 2 + 98 * 2) / 10,
        # buy_qty
        4,
        # sell_qty
        6,
        # buy_count
        2,
        # sell_count
        3
    ]))

    # add another after two seconds
    t7 = np.array([jh.now_to_timestamp() + 3000, 98, 2, 1], dtype=np.float64)
    store.trades.add_trade(t7, 'Sandbox', 'BTC-USD')

    np.testing.assert_equal(store.trades.get_current_trade('Sandbox', 'BTC-USD'), np.array([
        jh.now_to_timestamp() + 1000,
        # price
        98,
        # buy_qty
        2,
        # sell_qty
        0,
        # buy_count
        1,
        # sell_count
        0
    ]))

    # test get_past_trade
    np.testing.assert_equal(store.trades.get_past_trade('Sandbox', 'BTC-USD', 1), np.array([
        jh.now_to_timestamp(),
        # price
        (100 * 2 + 98 * 2 + 98 * 2 + 98 * 2 + 98 * 2) / 10,
        # buy_qty
        4,
        # sell_qty
        6,
        # buy_count
        2,
        # sell_count
        3
    ]))

    # test get_trades
    np.testing.assert_equal(store.trades.get_trades('Sandbox', 'BTC-USD'), np.array([
        [
            jh.now_to_timestamp(),
            # price
            (100 * 2 + 98 * 2 + 98 * 2 + 98 * 2 + 98 * 2) / 10,
            # buy_qty
            4,
            # sell_qty
            6,
            # buy_count
            2,
            # sell_count
            3
        ],
        [
            jh.now_to_timestamp() + 1000,
            # price
            98,
            # buy_qty
            2,
            # sell_qty
            0,
            # buy_count
            1,
            # sell_count
            0
        ]
    ]))
